#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import os
import urllib.parse

import pytest

import elasticapm
from elasticapm.conf import constants
from elasticapm.instrumentation.packages.botocore import SQS_MAX_ATTRIBUTES
from tests.utils import assert_any_record_contains

boto3 = pytest.importorskip("boto3")


pytestmark = [pytest.mark.boto3]

os.environ["AWS_ACCESS_KEY_ID"] = "key"
os.environ["AWS_SECRET_ACCESS_KEY"] = "secret"
os.environ["AWS_DEFAULT_REGION"] = "us-east-1"

LOCALSTACK_ENDPOINT = os.environ.get("AWS_URL", None)

from boto3.dynamodb.types import TypeSerializer

dynamodb_serializer = TypeSerializer()

if not LOCALSTACK_ENDPOINT:
    pytestmark.append(pytest.mark.skip("Skipping botocore tests, no AWS_URL environment variable set"))

LOCALSTACK_ENDPOINT_URL = urllib.parse.urlparse(LOCALSTACK_ENDPOINT)


@pytest.fixture()
def dynamodb():
    db = boto3.client("dynamodb", endpoint_url=LOCALSTACK_ENDPOINT)
    db.create_table(
        TableName="Movies",
        KeySchema=[
            {"AttributeName": "year", "KeyType": "HASH"},  # Partition key
            {"AttributeName": "title", "KeyType": "RANGE"},  # Sort key
        ],
        AttributeDefinitions=[
            {"AttributeName": "year", "AttributeType": "N"},
            {"AttributeName": "title", "AttributeType": "S"},
        ],
        ProvisionedThroughput={"ReadCapacityUnits": 10, "WriteCapacityUnits": 10},
    )
    yield db
    db.delete_table(TableName="Movies")


@pytest.fixture()
def sqs_client_and_queue():
    sqs = boto3.client("sqs", endpoint_url=LOCALSTACK_ENDPOINT)
    response = sqs.create_queue(QueueName="myqueue", Attributes={"MessageRetentionPeriod": "86400"})
    queue_url = response["QueueUrl"]
    yield sqs, queue_url
    sqs.delete_queue(QueueUrl=queue_url)


def test_botocore_instrumentation(instrument, elasticapm_client):
    elasticapm_client.begin_transaction("transaction.test")
    ec2 = boto3.client("ec2", endpoint_url=LOCALSTACK_ENDPOINT)
    ec2.describe_instances()
    elasticapm_client.end_transaction("MyView")
    span = elasticapm_client.events[constants.SPAN][0]

    assert span["name"] == "EC2:DescribeInstances"
    assert span["type"] == "aws"
    assert span["subtype"] == "ec2"
    assert span["action"] == "DescribeInstances"
    assert span["context"]["http"]["request"]["id"]


def test_s3(instrument, elasticapm_client):
    client = boto3.client("s3", endpoint_url=LOCALSTACK_ENDPOINT)
    elasticapm_client.begin_transaction("test")
    client.create_bucket(Bucket="xyz")
    client.put_object(Bucket="xyz", Key="abc", Body=b"foo")
    client.list_objects(Bucket="xyz")
    elasticapm_client.end_transaction("test", "test")
    transaction = elasticapm_client.end_transaction("test", "test")
    spans = elasticapm_client.events[constants.SPAN]
    for span in spans:
        assert span["type"] == "storage"
        assert span["subtype"] == "s3"
        assert span["context"]["destination"]["address"] == LOCALSTACK_ENDPOINT_URL.hostname
        assert span["context"]["destination"]["port"] == LOCALSTACK_ENDPOINT_URL.port
        assert span["context"]["destination"]["cloud"]["region"] == "us-east-1"
        assert span["context"]["destination"]["service"]["name"] == "s3"
        assert span["context"]["destination"]["service"]["resource"] == "s3/xyz"
        assert span["context"]["destination"]["service"]["type"] == "storage"
        assert span["context"]["http"]["request"]["id"]
    assert spans[0]["name"] == "S3 CreateBucket xyz"
    assert spans[0]["action"] == "CreateBucket"
    assert spans[1]["name"] == "S3 PutObject xyz"
    assert spans[1]["action"] == "PutObject"
    assert spans[1]["otel"]["attributes"]["aws.s3.bucket"] == "xyz"
    assert spans[1]["otel"]["attributes"]["aws.s3.key"] == "abc"
    assert spans[2]["name"] == "S3 ListObjects xyz"
    assert spans[2]["action"] == "ListObjects"


def test_dynamodb(instrument, elasticapm_client, dynamodb):
    elasticapm_client.begin_transaction("test")
    response = dynamodb.put_item(
        TableName="Movies",
        Item={"title": dynamodb_serializer.serialize("Independence Day"), "year": dynamodb_serializer.serialize(1994)},
    )
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200

    response = dynamodb.query(
        TableName="Movies",
        ExpressionAttributeValues={
            ":v1": {
                "S": "Independence Day",
            },
            ":v2": {
                "N": "1994",
            },
        },
        ExpressionAttributeNames={"#y": "year"},
        KeyConditionExpression="title = :v1 and #y = :v2",
    )
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200

    response = dynamodb.delete_item(
        TableName="Movies",
        Key={
            "title": {
                "S": '"Independence Day"',
            },
            "year": {
                "N": "1994",
            },
        },
    )
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200

    elasticapm_client.end_transaction("test", "test")
    spans = elasticapm_client.events[constants.SPAN]
    for span in spans:
        assert span["type"] == "db"
        assert span["subtype"] == "dynamodb"
        assert span["action"] == "query"
        assert span["context"]["db"]["instance"] == "us-east-1"
        assert span["context"]["db"]["type"] == "dynamodb"
        assert span["context"]["destination"]["address"] == LOCALSTACK_ENDPOINT_URL.hostname
        assert span["context"]["destination"]["port"] == LOCALSTACK_ENDPOINT_URL.port
        assert span["context"]["destination"]["cloud"]["region"] == "us-east-1"
        assert span["context"]["destination"]["service"]["name"] == "dynamodb"
        assert span["context"]["destination"]["service"]["resource"] == "Movies"
        assert span["context"]["destination"]["service"]["type"] == "db"
        assert span["context"]["http"]["request"]["id"]
    assert spans[0]["name"] == "DynamoDB PutItem Movies"
    assert spans[1]["name"] == "DynamoDB Query Movies"
    assert spans[1]["context"]["db"]["statement"] == "title = :v1 and #y = :v2"
    assert spans[2]["name"] == "DynamoDB DeleteItem Movies"


def test_sns(instrument, elasticapm_client):
    sns = boto3.client("sns", endpoint_url=LOCALSTACK_ENDPOINT)
    elasticapm_client.begin_transaction("test")
    response = sns.create_topic(Name="mytopic")
    topic_arn = response["TopicArn"]
    response = sns.list_topics()
    sns.publish(TopicArn=topic_arn, Subject="Saying", Message="this is my message to you-ou-ou")
    elasticapm_client.end_transaction("test", "test")
    spans = elasticapm_client.events[constants.SPAN]
    assert spans[2]["name"] == "SNS Publish mytopic"
    assert spans[2]["type"] == "messaging"
    assert spans[2]["subtype"] == "sns"
    assert spans[2]["action"] == "send"
    assert spans[2]["context"]["destination"]["address"] == LOCALSTACK_ENDPOINT_URL.hostname
    assert spans[2]["context"]["destination"]["port"] == LOCALSTACK_ENDPOINT_URL.port
    assert spans[2]["context"]["destination"]["cloud"]["region"] == "us-east-1"
    assert spans[2]["context"]["destination"]["service"]["name"] == "sns"
    assert spans[2]["context"]["destination"]["service"]["resource"] == "sns/mytopic"
    assert spans[2]["context"]["destination"]["service"]["type"] == "messaging"
    assert spans[2]["context"]["http"]["request"]["id"]


def test_sqs_send(instrument, elasticapm_client, sqs_client_and_queue):
    sqs, queue_url = sqs_client_and_queue
    elasticapm_client.begin_transaction("test")
    sqs.send_message(
        QueueUrl=queue_url,
        MessageAttributes={
            "Title": {"DataType": "String", "StringValue": "foo"},
        },
        MessageBody=("bar"),
    )
    transaction = elasticapm_client.end_transaction("test", "test")
    span = elasticapm_client.events[constants.SPAN][0]
    assert span["name"] == "SQS SEND to myqueue"
    assert span["type"] == "messaging"
    assert span["subtype"] == "sqs"
    assert span["action"] == "send"
    assert span["context"]["destination"]["cloud"]["region"] == "us-east-1"
    assert span["context"]["destination"]["service"]["name"] == "sqs"
    assert span["context"]["destination"]["service"]["resource"] == "sqs/myqueue"
    assert span["context"]["destination"]["service"]["type"] == "messaging"
    assert span["context"]["http"]["request"]["id"]

    messages = sqs.receive_message(
        QueueUrl=queue_url,
        AttributeNames=["All"],
        MessageAttributeNames=[
            "All",
        ],
    )
    message = messages["Messages"][0]
    assert "traceparent" in message["MessageAttributes"]
    traceparent = message["MessageAttributes"]["traceparent"]["StringValue"]
    assert transaction.trace_parent.trace_id in traceparent
    assert span["id"] in traceparent


def test_sqs_send_batch(instrument, elasticapm_client, sqs_client_and_queue):
    sqs, queue_url = sqs_client_and_queue
    elasticapm_client.begin_transaction("test")
    response = sqs.send_message_batch(
        QueueUrl=queue_url,
        Entries=[
            {
                "Id": "foo",
                "MessageBody": "foo",
                "MessageAttributes": {"string": {"StringValue": "foo", "DataType": "String"}},
            },
        ],
    )
    transaction = elasticapm_client.end_transaction("test", "test")
    span = elasticapm_client.events[constants.SPAN][0]
    assert span["name"] == "SQS SEND_BATCH to myqueue"
    assert span["type"] == "messaging"
    assert span["subtype"] == "sqs"
    assert span["action"] == "send_batch"
    assert span["context"]["destination"]["cloud"]["region"] == "us-east-1"
    assert span["context"]["destination"]["service"]["name"] == "sqs"
    assert span["context"]["destination"]["service"]["resource"] == "sqs/myqueue"
    assert span["context"]["destination"]["service"]["type"] == "messaging"
    messages = sqs.receive_message(
        QueueUrl=queue_url,
        AttributeNames=["All"],
        MessageAttributeNames=[
            "All",
        ],
    )
    message = messages["Messages"][0]
    assert "traceparent" in message["MessageAttributes"]
    traceparent = message["MessageAttributes"]["traceparent"]["StringValue"]
    assert transaction.trace_parent.trace_id in traceparent
    assert span["id"] in traceparent


def test_sqs_no_message_attributes(instrument, elasticapm_client, sqs_client_and_queue):
    sqs, queue_url = sqs_client_and_queue
    elasticapm_client.begin_transaction("test")
    response = sqs.send_message_batch(
        QueueUrl=queue_url,
        Entries=[
            {
                "Id": "foo",
                "MessageBody": "foo",
            },
        ],
    )
    transaction = elasticapm_client.end_transaction("test", "test")
    span = elasticapm_client.events[constants.SPAN][0]
    messages = sqs.receive_message(
        QueueUrl=queue_url,
        AttributeNames=["All"],
        MessageAttributeNames=[
            "All",
        ],
    )
    message = messages["Messages"][0]
    assert "traceparent" in message["MessageAttributes"]
    traceparent = message["MessageAttributes"]["traceparent"]["StringValue"]
    assert transaction.trace_parent.trace_id in traceparent
    assert span["id"] in traceparent


def test_sqs_send_too_many_attributes_for_disttracing(instrument, elasticapm_client, sqs_client_and_queue, caplog):
    sqs, queue_url = sqs_client_and_queue
    attributes = {str(i): {"DataType": "String", "StringValue": str(i)} for i in range(SQS_MAX_ATTRIBUTES)}
    elasticapm_client.begin_transaction("test")
    with caplog.at_level("INFO"):
        sqs.send_message(
            QueueUrl=queue_url,
            MessageAttributes=attributes,
            MessageBody=("bar"),
        )
    elasticapm_client.end_transaction("test", "test")
    messages = sqs.receive_message(
        QueueUrl=queue_url,
        AttributeNames=["All"],
        MessageAttributeNames=[
            "All",
        ],
    )
    message = messages["Messages"][0]
    assert "traceparent" not in message["MessageAttributes"]
    assert_any_record_contains(caplog.records, "Not adding disttracing headers")


def test_sqs_send_disttracing_dropped_span(instrument, elasticapm_client, sqs_client_and_queue):
    sqs, queue_url = sqs_client_and_queue
    elasticapm_client.begin_transaction("test")
    with elasticapm.capture_span("test", leaf=True):
        sqs.send_message(
            QueueUrl=queue_url,
            MessageAttributes={
                "Title": {"DataType": "String", "StringValue": "foo"},
            },
            MessageBody=("bar"),
        )
    transaction = elasticapm_client.end_transaction("test", "test")
    assert len(elasticapm_client.events[constants.SPAN]) == 1
    messages = sqs.receive_message(
        QueueUrl=queue_url,
        AttributeNames=["All"],
        MessageAttributeNames=[
            "All",
        ],
    )
    message = messages["Messages"][0]
    assert "traceparent" in message["MessageAttributes"]
    traceparent = message["MessageAttributes"]["traceparent"]["StringValue"]
    assert transaction.trace_parent.trace_id in traceparent
    assert transaction.id in traceparent  # due to DroppedSpan, transaction.id is used instead of span.id


def test_sqs_receive_and_delete(instrument, elasticapm_client, sqs_client_and_queue):
    sqs, queue_url = sqs_client_and_queue
    sqs.send_message(
        QueueUrl=queue_url,
        MessageAttributes={
            "Title": {"DataType": "String", "StringValue": "foo"},
        },
        MessageBody=("bar"),
    )
    elasticapm_client.begin_transaction("test")
    response = sqs.receive_message(
        QueueUrl=queue_url,
        AttributeNames=["All"],
        MessageAttributeNames=[
            "All",
        ],
    )
    sqs.delete_message(QueueUrl=queue_url, ReceiptHandle=response["Messages"][0]["ReceiptHandle"])
    elasticapm_client.end_transaction("test", "test")

    receive_span = elasticapm_client.events[constants.SPAN][0]
    assert receive_span["name"] == "SQS RECEIVE from myqueue"
    assert receive_span["type"] == "messaging"
    assert receive_span["subtype"] == "sqs"
    assert receive_span["action"] == "receive"
    assert receive_span["context"]["destination"]["cloud"]["region"] == "us-east-1"
    assert receive_span["context"]["destination"]["service"]["name"] == "sqs"
    assert receive_span["context"]["destination"]["service"]["resource"] == "sqs/myqueue"
    assert receive_span["context"]["destination"]["service"]["type"] == "messaging"

    delete_span = elasticapm_client.events[constants.SPAN][1]
    assert delete_span["name"] == "SQS DELETE from myqueue"
    assert delete_span["type"] == "messaging"
    assert delete_span["subtype"] == "sqs"
    assert delete_span["action"] == "delete"
    assert delete_span["context"]["destination"]["cloud"]["region"] == "us-east-1"
    assert delete_span["context"]["destination"]["service"]["name"] == "sqs"
    assert delete_span["context"]["destination"]["service"]["resource"] == "sqs/myqueue"
    assert delete_span["context"]["destination"]["service"]["type"] == "messaging"


def test_sqs_delete_batch(instrument, elasticapm_client, sqs_client_and_queue):
    sqs, queue_url = sqs_client_and_queue
    sqs.send_message(
        QueueUrl=queue_url,
        MessageAttributes={
            "Title": {"DataType": "String", "StringValue": "foo"},
        },
        MessageBody=("bar"),
    )
    response = sqs.receive_message(
        QueueUrl=queue_url,
        AttributeNames=["All"],
        MessageAttributeNames=[
            "All",
        ],
    )
    elasticapm_client.begin_transaction("test")
    sqs.delete_message_batch(
        QueueUrl=queue_url,
        Entries=[{"Id": "foo", "ReceiptHandle": response["Messages"][0]["ReceiptHandle"]}],
    )
    elasticapm_client.end_transaction("test", "test")

    delete_span = elasticapm_client.events[constants.SPAN][0]
    assert delete_span["name"] == "SQS DELETE_BATCH from myqueue"
    assert delete_span["type"] == "messaging"
    assert delete_span["subtype"] == "sqs"
    assert delete_span["action"] == "delete_batch"
    assert delete_span["context"]["destination"]["cloud"]["region"] == "us-east-1"
    assert delete_span["context"]["destination"]["service"]["name"] == "sqs"
    assert delete_span["context"]["destination"]["service"]["resource"] == "sqs/myqueue"
    assert delete_span["context"]["destination"]["service"]["type"] == "messaging"


def test_sqs_receive_message_span_links(instrument, elasticapm_client, sqs_client_and_queue):
    sqs, queue_url = sqs_client_and_queue
    send_transaction = elasticapm_client.begin_transaction("test")
    sqs.send_message(
        QueueUrl=queue_url,
        MessageBody=("bar"),
    )
    elasticapm_client.end_transaction("test")
    receive_transaction = elasticapm_client.begin_transaction("test")
    response = sqs.receive_message(
        QueueUrl=queue_url,
        AttributeNames=["All"],
    )
    assert len(response["Messages"]) == 1
    elasticapm_client.end_transaction("test")
    send_span = elasticapm_client.events[constants.SPAN][0]
    receive_span = elasticapm_client.events[constants.SPAN][1]
    assert receive_span["links"][0]["trace_id"] == send_transaction.trace_parent.trace_id
    assert receive_span["links"][0]["span_id"] == send_span["id"]



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import pytest

from elasticapm.conf.constants import TRANSACTION
from elasticapm.traces import capture_span

graphene = pytest.importorskip("graphene")
graphql = pytest.importorskip("graphql")

pytestmark = pytest.mark.graphene


class Success(graphene.ObjectType):
    yeah = graphene.String(required=True)


class Error(graphene.ObjectType):
    message = graphene.String()


class CreatePostResult(graphene.Union):
    class Meta:
        types = [Success, Error]


class CreatePost(graphene.Mutation):
    class Arguments:
        text = graphene.String(required=True)

    result = graphene.Field(CreatePostResult)

    def mutate(self, info, text):
        result = Success(yeah="yeah")

        return CreatePost(result=result)


class Mutations(graphene.ObjectType):
    create_post = CreatePost.Field()


class Query(graphene.ObjectType):
    succ = graphene.Field(Success)
    err = graphene.Field(Error)
    gevent = graphene.Field(Success)

    def resolve_succ(self, *args, **kwargs):
        return Success(yeah="hello world")

    def resolve_err(self, *args, **kwargs):
        #        import pdb; pdb.set_trace()
        return Error(message="oops")


@pytest.mark.skipif(
    not hasattr(graphql, "VERSION") or graphql.VERSION[0] >= 3, reason="Executor is reimplementated in graphql-core 3"
)
@pytest.mark.integrationtest
def test_create_post(instrument, elasticapm_client):
    query_string = """
    mutation {
      createPost(text: "Try this out") {
        result {
          __typename
        }
      }
    }
    """

    schema = graphene.Schema(query=Query, mutation=Mutations)

    elasticapm_client.begin_transaction("transaction.test")
    with capture_span("test_graphene", "test"):
        result = schema.execute(query_string)
        assert not result.errors
        assert result.data["createPost"]["result"]["__typename"] == "Success"
    elasticapm_client.end_transaction("BillingView")
    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])
    expected_signatures = {
        "GraphQL.mutation createPost",
        "GraphQL.mutation result",
        "test_graphene",
    }
    assert {t["name"] for t in spans} == expected_signatures
    assert transactions[0]["name"] == "GraphQL MUTATION createPost"


@pytest.mark.skipif(
    not hasattr(graphql, "VERSION") or graphql.VERSION[0] >= 3, reason="Executor is reimplementated in graphql-core 3"
)
@pytest.mark.integrationtest
def test_fetch_data(instrument, elasticapm_client):
    query_string = "{succ{yeah},err{__typename}}"

    schema = graphene.Schema(query=Query)

    elasticapm_client.begin_transaction("transaction.test")
    with capture_span("test_graphene", "test"):
        result = schema.execute(query_string)
        assert result.data == {"succ": {"yeah": "hello world"}, "err": {"__typename": "Error"}}
    elasticapm_client.end_transaction("BillingView")
    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])
    expected_signatures = {
        "GraphQL.query err",
        "GraphQL.query succ",
        "test_graphene",
    }
    assert {t["name"] for t in spans} == expected_signatures
    assert transactions[0]["name"] == "GraphQL QUERY succ+err"



# -*- coding: utf-8 -*-

#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import datetime
import decimal
import uuid

import pytest

simplejson = pytest.importorskip("simplejson")

from elasticapm.utils import simplejson_encoder as json


def test_uuid():
    res = uuid.uuid4()
    assert json.dumps(res) == '"%s"' % res.hex


def test_datetime():
    res = datetime.datetime(day=1, month=1, year=2011, hour=1, minute=1, second=1)
    assert json.dumps(res) == '"2011-01-01T01:01:01.000000Z"'


def test_set():
    res = set(["foo", "bar"])
    assert json.dumps(res) in ('["foo", "bar"]', '["bar", "foo"]')


def test_frozenset():
    res = frozenset(["foo", "bar"])
    assert json.dumps(res) in ('["foo", "bar"]', '["bar", "foo"]')


def test_bytes():
    res = bytes("foobar", encoding="ascii")
    assert json.dumps(res) == '"foobar"'


def test_decimal():
    res = decimal.Decimal("1.0")
    assert json.dumps(res) == "1.0"


@pytest.mark.parametrize("res", [float("nan"), float("+inf"), float("-inf")])
def test_float_invalid_json(res):
    assert json.dumps(res) == "null"


def test_float():
    res = 1.0
    assert json.dumps(res) == "1.0"


def test_unsupported():
    res = object()
    assert json.dumps(res).startswith('"<object object at')



#  BSD 3-Clause License
#
#  Copyright (c) 2022, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import asyncio
import logging
import os
import sys
from concurrent import futures

import grpc

import elasticapm
from elasticapm.contrib.grpc import GRPCApmClient

from . import testgrpc_pb2 as pb2
from . import testgrpc_pb2_grpc as pb2_grpc

elasticapm.instrument()


class TestService(pb2_grpc.TestServiceServicer):
    def __init__(self, *args, **kwargs) -> None:
        pass

    def GetServerResponse(self, request, context):
        message = request.message
        result = f'Hello I am up and running received "{message}" message from you'
        result = {"message": result, "received": True}

        return pb2.MessageResponse(**result)

    def GetServerResponseAbort(self, request, context) -> None:
        context.abort(grpc.StatusCode.INTERNAL, "foo")

    def GetServerResponseUnavailable(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNAVAILABLE)
        context.set_details("Method not available")
        return pb2.MessageResponse(message="foo", received=True)

    def GetServerResponseException(self, request, context):
        raise Exception("oh no")


class TestServiceAsync(pb2_grpc.TestServiceServicer):
    def __init__(self, *args, **kwargs) -> None:
        pass

    async def GetServerResponse(self, request, context):
        message = request.message
        result = f'Hello I am up and running received "{message}" message from you'
        result = {"message": result, "received": True}

        return pb2.MessageResponse(**result)

    async def GetServerResponseAbort(self, request, context) -> None:
        await context.abort(grpc.StatusCode.INTERNAL, "foo")

    async def GetServerResponseUnavailable(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNAVAILABLE)
        context.set_details("Method not available")
        return pb2.MessageResponse(message="foo", received=True)

    async def GetServerResponseException(self, request, context):
        raise Exception("oh no")


def serve(port) -> None:
    apm_client = GRPCApmClient(
        service_name="grpc-server", disable_metrics="*", api_request_time="100ms", central_config="False"
    )
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
    pb2_grpc.add_TestServiceServicer_to_server(TestService(), server)
    server.add_insecure_port(f"[::]:{port}")
    server.start()
    server.wait_for_termination()


async def serve_async(port) -> None:
    apm_client = GRPCApmClient(
        service_name="grpc-server", disable_metrics="*", api_request_time="100ms", central_config="False"
    )
    server = grpc.aio.server()
    pb2_grpc.add_TestServiceServicer_to_server(TestServiceAsync(), server)
    server.add_insecure_port(f"[::]:{port}")
    await server.start()
    await server.wait_for_termination()


if __name__ == "__main__":
    if len(sys.argv) > 1:
        port = sys.argv[1]
    else:
        port = "50051"
    logging.basicConfig()
    if os.environ.get("GRPC_SERVER_ASYNC") == "1":
        asyncio.run(serve_async(port))
    else:
        serve(port)



#  BSD 3-Clause License
#
#  Copyright (c) 2012, the Sentry Team, see AUTHORS for more details
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE


import fnmatch
import inspect
import itertools
import os
import re
import sys
from functools import lru_cache

from elasticapm.utils.encoding import transform

_coding_re = re.compile(r"coding[:=]\s*([-\w.]+)")


@lru_cache(512)
def get_lines_from_file(filename, lineno, context_lines, loader=None, module_name=None):
    """
    Returns context_lines before and after lineno from file.
    Returns (pre_context_lineno, pre_context, context_line, post_context).
    """
    lineno = lineno - 1
    lower_bound = max(0, lineno - context_lines)
    upper_bound = lineno + context_lines

    source = None
    if loader is not None and hasattr(loader, "get_source"):
        result = get_source_lines_from_loader(loader, module_name, lineno, lower_bound, upper_bound)
        if result is not None:
            return result

    if source is None:
        try:
            with open(filename, "rb") as file_obj:
                encoding = "utf8"
                # try to find encoding of source file by "coding" header
                # if none is found, utf8 is used as a fallback
                for line in itertools.islice(file_obj, 0, 2):
                    match = _coding_re.search(line.decode("utf8"))
                    if match:
                        encoding = match.group(1)
                        break
                file_obj.seek(0)
                lines = [
                    str(line, encoding, "replace") for line in itertools.islice(file_obj, lower_bound, upper_bound + 1)
                ]
                offset = lineno - lower_bound
                return (
                    [line.strip("\r\n") for line in lines[0:offset]],
                    lines[offset].strip("\r\n"),
                    [line.strip("\r\n") for line in lines[offset + 1 :]] if len(lines) > offset else [],
                )
        except (OSError, IOError, IndexError):
            pass
    return None, None, None


def get_source_lines_from_loader(loader, module_name, lineno, lower_bound, upper_bound):
    try:
        source = loader.get_source(module_name)
    except ImportError:
        # ImportError: Loader for module cProfile cannot handle module __main__
        return None
    if source is not None:
        source = source.splitlines()
    else:
        return None, None, None
    try:
        pre_context = [line.strip("\r\n") for line in source[lower_bound:lineno]]
        context_line = source[lineno].strip("\r\n")
        post_context = [line.strip("\r\n") for line in source[(lineno + 1) : upper_bound + 1]]
    except IndexError:
        # the file may have changed since it was loaded into memory
        return None, None, None
    return pre_context, context_line, post_context


def get_culprit(frames, include_paths=None, exclude_paths=None):
    # We iterate through each frame looking for a deterministic culprit
    # When one is found, we mark it as last "best guess" (best_guess) and then
    # check it against ``exclude_paths``. If it isnt listed, then we
    # use this option. If nothing is found, we use the "best guess".
    if include_paths is None:
        include_paths = []
    if exclude_paths is None:
        exclude_paths = []
    best_guess = None
    culprit = None
    for frame in frames:
        try:
            culprit = ".".join((f or "<unknown>" for f in [frame.get("module"), frame.get("function")]))
        except KeyError:
            continue
        if any((culprit.startswith(k) for k in include_paths)):
            if not (best_guess and any((culprit.startswith(k) for k in exclude_paths))):
                best_guess = culprit
        elif best_guess:
            break

    # Return either the best guess or the last frames call
    return best_guess or culprit


def _getitem_from_frame(f_locals, key, default=None):
    """
    f_locals is not guaranteed to have .get(), but it will always
    support __getitem__. Even if it doesnt, we return ``default``.
    """
    try:
        return f_locals[key]
    except Exception:
        return default


def to_dict(dictish):
    """
    Given something that closely resembles a dictionary, we attempt
    to coerce it into a propery dictionary.
    """
    if hasattr(dictish, "iterkeys"):
        m = dictish.iterkeys
    elif hasattr(dictish, "keys"):
        m = dictish.keys
    else:
        raise ValueError(dictish)

    return dict((k, dictish[k]) for k in m())


def iter_traceback_frames(tb, config=None):
    """
    Given a traceback object, it will iterate over all
    frames that do not contain the ``__traceback_hide__``
    local variable.
    """
    max_frames = config.stack_trace_limit if config else -1
    if not max_frames:
        return
    frames = []
    while tb:
        # support for __traceback_hide__ which is used by a few libraries
        # to hide internal frames.
        frame = tb.tb_frame
        f_locals = getattr(frame, "f_locals", None)
        if f_locals is None or not _getitem_from_frame(f_locals, "__traceback_hide__"):
            frames.append((frame, getattr(tb, "tb_lineno", None)))
        tb = tb.tb_next
    if max_frames != -1:
        frames = frames[-max_frames:]
    for frame in frames:
        yield frame


def iter_stack_frames(frames=None, start_frame=None, skip=0, skip_top_modules=(), config=None):
    """
    Given an optional list of frames (defaults to current stack),
    iterates over all frames that do not contain the ``__traceback_hide__``
    local variable.

    Frames can be skipped by either providing a number, or a tuple
    of module names. If the module of a frame matches one of the names
    (using `.startswith`, that frame will be skipped. This matching will only
    be done until the first frame doesn't match.

    This is useful to filter out frames that are caused by frame collection
    itself.

    :param frames: a list of frames, or None
    :param start_frame: a Frame object or None
    :param skip: number of frames to skip from the beginning
    :param skip_top_modules: tuple of strings
    :param config: agent configuration

    """
    if not frames:
        frame = start_frame if start_frame is not None else inspect.currentframe().f_back
        frames = _walk_stack(frame)
    max_frames = config.stack_trace_limit if config else -1
    stop_ignoring = False
    frames_count = 0  # can't use i, as we don't want to count skipped and ignored frames
    for i, frame in enumerate(frames):
        if max_frames != -1 and frames_count == max_frames:
            break
        if i < skip:
            continue
        f_globals = getattr(frame, "f_globals", {})
        if not stop_ignoring and f_globals.get("__name__", "").startswith(skip_top_modules):
            continue
        stop_ignoring = True
        f_locals = getattr(frame, "f_locals", {})
        if not _getitem_from_frame(f_locals, "__traceback_hide__"):
            frames_count += 1
            yield frame, frame.f_lineno


def get_frame_info(
    frame,
    lineno,
    with_locals=True,
    library_frame_context_lines=None,
    in_app_frame_context_lines=None,
    include_paths_re=None,
    exclude_paths_re=None,
    locals_processor_func=None,
):
    # Support hidden frames
    f_locals = getattr(frame, "f_locals", {})
    if _getitem_from_frame(f_locals, "__traceback_hide__"):
        return None

    f_globals = getattr(frame, "f_globals", {})
    loader = f_globals.get("__loader__")
    module_name = f_globals.get("__name__")

    f_code = getattr(frame, "f_code", None)
    if f_code:
        abs_path = frame.f_code.co_filename
        function = frame.f_code.co_name
    else:
        abs_path = None
        function = None

    # Try to pull a relative file path
    # This changes /foo/site-packages/baz/bar.py into baz/bar.py
    try:
        base_filename = sys.modules[module_name.split(".", 1)[0]].__file__
        filename = abs_path.split(base_filename.rsplit(os.path.sep, 2)[0], 1)[-1].lstrip(os.path.sep)
    except Exception:
        filename = abs_path

    if not filename:
        filename = abs_path

    frame_result = {
        "abs_path": abs_path,
        "filename": filename,
        "module": module_name,
        "function": function,
        "lineno": lineno,
        "library_frame": is_library_frame(abs_path, include_paths_re, exclude_paths_re),
    }

    context_lines = library_frame_context_lines if frame_result["library_frame"] else in_app_frame_context_lines
    if context_lines and lineno is not None and abs_path:
        # context_metadata will be processed by elasticapm.processors.add_context_lines_to_frames.
        # This ensures that blocking operations (reading from source files) happens on the background
        # processing thread.
        frame_result["context_metadata"] = (abs_path, lineno, int(context_lines / 2), loader, module_name)
    if with_locals:
        if f_locals is not None and not isinstance(f_locals, dict):
            # XXX: Genshi (and maybe others) have broken implementations of
            # f_locals that are not actually dictionaries
            try:
                f_locals = to_dict(f_locals)
            except Exception:
                f_locals = "<invalid local scope>"
        if locals_processor_func:
            f_locals = {varname: locals_processor_func(var) for varname, var in f_locals.items()}
        frame_result["vars"] = transform(f_locals)
    return frame_result


def get_stack_info(
    frames,
    with_locals=True,
    library_frame_context_lines=None,
    in_app_frame_context_lines=None,
    include_paths_re=None,
    exclude_paths_re=None,
    locals_processor_func=None,
):
    """
    Given a list of frames, returns a list of stack information
    dictionary objects that are JSON-ready.

    We have to be careful here as certain implementations of the
    _Frame class do not contain the necessary data to lookup all
    of the information we want.

    :param frames: a list of (Frame, lineno) tuples
    :param with_locals: boolean to indicate if local variables should be collected
    :param include_paths_re: a regex to determine if a frame is not a library frame
    :param exclude_paths_re: a regex to exclude frames from not being library frames
    :param locals_processor_func: a function to call on all local variables
    :return:
    """
    results = []
    for frame, lineno in frames:
        result = get_frame_info(
            frame,
            lineno,
            library_frame_context_lines=library_frame_context_lines,
            in_app_frame_context_lines=in_app_frame_context_lines,
            with_locals=with_locals,
            include_paths_re=include_paths_re,
            exclude_paths_re=exclude_paths_re,
            locals_processor_func=locals_processor_func,
        )
        if result:
            results.append(result)
    return results


def _walk_stack(frame):
    while frame:
        yield frame
        frame = frame.f_back


@lru_cache(512)
def is_library_frame(abs_file_path, include_paths_re, exclude_paths_re):
    if not abs_file_path:
        return True
    if include_paths_re and include_paths_re.match(abs_file_path):
        # frame is in-app, return False
        return False
    elif exclude_paths_re and exclude_paths_re.match(abs_file_path):
        return True
    # if neither excluded nor included, assume it's an in-app frame
    return False


def get_path_regex(paths):
    """
    compiles a list of path globs into a single pattern that matches any of the given paths
    :param paths: a list of strings representing fnmatch path globs
    :return: a compiled regex
    """
    return re.compile("|".join(map(fnmatch.translate, paths)))



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import io

import pytest

from elasticapm.utils import cgroup


@pytest.mark.parametrize(
    "test_input,expected",
    [
        (
            "12:devices:/docker/051e2ee0bce99116029a13df4a9e943137f19f957f38ac02d6bad96f9b700f76",
            {"container": {"id": "051e2ee0bce99116029a13df4a9e943137f19f957f38ac02d6bad96f9b700f76"}},
        ),
        (
            "1:name=systemd:/system.slice/docker-cde7c2bab394630a42d73dc610b9c57415dced996106665d427f6d0566594411.scope",
            {"container": {"id": "cde7c2bab394630a42d73dc610b9c57415dced996106665d427f6d0566594411"}},
        ),
        (
            "1:name=systemd:/kubepods/besteffort/pode9b90526-f47d-11e8-b2a5-080027b9f4fb/15aa6e53-b09a-40c7-8558-c6c31e36c88a",
            {
                "container": {"id": "15aa6e53-b09a-40c7-8558-c6c31e36c88a"},
                "kubernetes": {"pod": {"uid": "e9b90526-f47d-11e8-b2a5-080027b9f4fb"}},
            },
        ),
        (
            "1:name=systemd:/kubepods.slice/kubepods-burstable.slice/kubepods-burstable-pod90d81341_92de_11e7_8cf2_507b9d4141fa.slice/crio-2227daf62df6694645fee5df53c1f91271546a9560e8600a525690ae252b7f63.scope",
            {
                "container": {"id": "2227daf62df6694645fee5df53c1f91271546a9560e8600a525690ae252b7f63"},
                "kubernetes": {"pod": {"uid": "90d81341-92de-11e7-8cf2-507b9d4141fa"}},
            },
        ),
        (
            "1:name=systemd:/system.slice/garden.service/garden/70eb4ce5-a065-4401-6990-88ed",
            {"container": {"id": "70eb4ce5-a065-4401-6990-88ed"}},
        ),
        (
            "12:pids:/kubepods/kubepods/besteffort/pod0e886e9a-3879-45f9-b44d-86ef9df03224/244a65edefdffe31685c42317c9054e71dc1193048cf9459e2a4dd35cbc1dba4",
            {
                "container": {"id": "244a65edefdffe31685c42317c9054e71dc1193048cf9459e2a4dd35cbc1dba4"},
                "kubernetes": {"pod": {"uid": "0e886e9a-3879-45f9-b44d-86ef9df03224"}},
            },
        ),
        (
            "10:cpuset:/kubepods/pod5eadac96-ab58-11ea-b82b-0242ac110009/7fe41c8a2d1da09420117894f11dd91f6c3a44dfeb7d125dc594bd53468861df",
            {
                "container": {"id": "7fe41c8a2d1da09420117894f11dd91f6c3a44dfeb7d125dc594bd53468861df"},
                "kubernetes": {"pod": {"uid": "5eadac96-ab58-11ea-b82b-0242ac110009"}},
            },
        ),
        (
            "9:freezer:/kubepods.slice/kubepods-pod22949dce_fd8b_11ea_8ede_98f2b32c645c.slice/docker-b15a5bdedd2e7645c3be271364324321b908314e4c77857bbfd32a041148c07f.scope",
            {
                "container": {"id": "b15a5bdedd2e7645c3be271364324321b908314e4c77857bbfd32a041148c07f"},
                "kubernetes": {"pod": {"uid": "22949dce-fd8b-11ea-8ede-98f2b32c645c"}},
            },
        ),
    ],
)
def test_cgroup_parsing(test_input, expected):
    f = io.StringIO(test_input)
    result = cgroup.parse_cgroups(f)
    assert result == expected



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import time

import pytest

from elasticapm.metrics.base_metrics import MetricsRegistry

cpu_psutil = pytest.importorskip("elasticapm.metrics.sets.cpu_psutil")
pytestmark = pytest.mark.psutil


def test_cpu_mem_from_psutil(elasticapm_client):
    metricset = cpu_psutil.CPUMetricSet(MetricsRegistry(elasticapm_client))
    # do something that generates some CPU load
    for i in range(10**6):
        j = i * i
    data = next(metricset.collect())
    # we can't really test any specific values here as it depends on the system state.
    # Mocking is also not really a viable choice, as we would then lose the "integration testing"
    # nature of this test with different versions of psutil
    # Please note that cpu percentage may be > 1 on multi-threading applications using more
    # than one core
    assert data["samples"]["system.cpu.total.norm.pct"]["value"] > 0
    assert data["samples"]["system.process.cpu.total.norm.pct"]["value"] > 0

    assert data["samples"]["system.memory.total"]["value"] > 0
    assert data["samples"]["system.memory.actual.free"]["value"] > 0

    assert data["samples"]["system.process.memory.rss.bytes"]["value"] > 0
    assert data["samples"]["system.process.memory.size"]["value"] > 0


cpu_linux = pytest.importorskip("elasticapm.metrics.sets.cpu_linux")


@pytest.mark.skip("test is flaky on CI")
def test_compare_psutil_linux_metricsets(elasticapm_client):
    psutil_metricset = cpu_psutil.CPUMetricSet(MetricsRegistry(elasticapm_client))
    linux_metricset = cpu_linux.CPUMetricSet(MetricsRegistry(elasticapm_client))
    # do something that generates some CPU load
    for i in range(10**6):
        j = i * i
    psutil_data = next(psutil_metricset.collect())
    linux_data = next(linux_metricset.collect())

    assert (
        abs(
            psutil_data["samples"]["system.cpu.total.norm.pct"]["value"]
            - linux_data["samples"]["system.cpu.total.norm.pct"]["value"]
        )
        < 0.02
    )
    assert (
        abs(
            psutil_data["samples"]["system.process.cpu.total.norm.pct"]["value"]
            - linux_data["samples"]["system.process.cpu.total.norm.pct"]["value"]
        )
        < 0.02
    )



#  BSD 3-Clause License
#
#  Copyright (c) 2022, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging

from elasticapm.instrumentation.packages.base import AbstractInstrumentedModule


class FlaskInstrumentation(AbstractInstrumentedModule):
    name = "flask"

    instrument_list = [("flask", "Flask.__init__")]

    creates_transactions = True

    def call(self, module, method, wrapped, instance, args, kwargs) -> None:
        from elasticapm.contrib.flask import ElasticAPM

        wrapped(*args, **kwargs)
        client = ElasticAPM(instance)
        instance.elasticapm_client = client
        self.instance = instance

    def uninstrument(self) -> None:
        """
        This is mostly here for testing. If we want to support live
        instrumenting and uninstrumenting, we'll need to also extend the
        `instrument()` method to add the signals removed here.
        """
        super().uninstrument()

        # Only remove signals during uninstrument if we auto-instrumented
        flask_app = getattr(self, "instance", None)
        if flask_app:
            client = flask_app.elasticapm_client
            from flask import signals

            signals.request_started.disconnect(client.request_started)
            signals.request_finished.disconnect(client.request_finished)
            # remove logging handler if it was added
            logger = logging.getLogger()
            for handler in list(logger.handlers):
                if getattr(handler, "client", None) is client.client:
                    logger.removeHandler(handler)
            self.instance = None



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os

import pytest
import pytest_asyncio

from elasticapm.conf import constants

asyncpg = pytest.importorskip("asyncpg")  # isort:skip
pytestmark = [pytest.mark.asyncpg, pytest.mark.asyncio]

if "POSTGRES_DB" not in os.environ:
    pytestmark.append(pytest.mark.skip("Skipping asyncpg tests, no POSTGRES_DB environment variable set"))


def dsn():
    return "postgres://{user}:{password}@{host}:{port}/{database}".format(
        **{
            "database": os.environ.get("POSTGRES_DB", "elasticapm_test"),
            "user": os.environ.get("POSTGRES_USER", "postgres"),
            "password": os.environ.get("POSTGRES_PASSWORD", "postgres"),
            "host": os.environ.get("POSTGRES_HOST", "localhost"),
            "port": os.environ.get("POSTGRES_PORT", "5432"),
        }
    )


@pytest_asyncio.fixture()
async def connection(request):
    conn = await asyncpg.connect(dsn())

    await conn.execute(
        "BEGIN;"
        "CREATE TABLE test(id int, name VARCHAR(5) NOT NULL);"
        "INSERT INTO test VALUES (1, 'one'), (2, 'two'), (3, 'three');"
    )
    yield conn

    await conn.execute("ROLLBACK")
    await conn.close()


async def test_execute_with_sleep(instrument, connection, elasticapm_client):
    elasticapm_client.begin_transaction("test")
    await connection.execute("SELECT pg_sleep(0.1);")
    elasticapm_client.end_transaction("test", "OK")

    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)

    assert len(spans) == 1
    span = spans[0]
    assert 100 < span["duration"] < 110
    assert transaction["id"] == span["transaction_id"]
    assert span["type"] == "db"
    assert span["subtype"] == "postgresql"
    assert span["action"] == "query"
    assert span["sync"] == False
    assert span["name"] == "SELECT FROM"


async def test_executemany(instrument, connection, elasticapm_client):
    elasticapm_client.begin_transaction("test")
    await connection.executemany("INSERT INTO test VALUES ($1, $2);", [(1, "uno"), (2, "due")])
    elasticapm_client.end_transaction("test", "OK")

    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)

    assert len(spans) == 1
    span = spans[0]
    assert transaction["id"] == span["transaction_id"]
    assert span["subtype"] == "postgresql"
    assert span["action"] == "query"
    assert span["sync"] == False
    assert span["name"] == "INSERT INTO test"


def _assert_fetch(result):
    assert len(result) == 3
    assert all(isinstance(val, asyncpg.Record) for val in result)


def _assert_fetchval(result):
    assert result == 1


def _assert_fetchrow(result):
    assert isinstance(result, asyncpg.Record)
    assert result["id"] == 1


@pytest.mark.usefixtures("instrument")
@pytest.mark.parametrize(
    "method,verify", [("fetch", _assert_fetch), ("fetchval", _assert_fetchval), ("fetchrow", _assert_fetchrow)]
)
async def test_fetch_methods(connection, elasticapm_client, method, verify):
    elasticapm_client.begin_transaction("test")
    result = await getattr(connection, method)("SELECT id FROM test;")
    elasticapm_client.end_transaction("test", "OK")

    verify(result)

    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)

    assert len(spans) == 1
    span = spans[0]
    assert transaction["id"] == span["transaction_id"]
    assert span["subtype"] == "postgresql"
    assert span["action"] == "query"
    assert span["sync"] is False
    assert span["name"] == "SELECT FROM test"


@pytest.mark.usefixtures("instrument")
async def test_truncate_long_sql(connection, elasticapm_client):
    elasticapm_client.begin_transaction("test")
    await connection.execute(f"SELECT id, name FROM test WHERE name = '{'x' * 10010}';")
    elasticapm_client.end_transaction("test", "OK")

    transactions = elasticapm_client.events[constants.TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])

    statement = spans[0]["context"]["db"]["statement"]
    assert len(statement) == 10000
    assert statement.endswith("...")



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from tests.fixtures import TempStoreClient

import pytest  # isort:skip

fastapi = pytest.importorskip("fastapi")  # isort:skip

from fastapi import APIRouter, FastAPI
from starlette.applications import Starlette
from starlette.responses import PlainTextResponse
from starlette.routing import Route
from starlette.testclient import TestClient

import elasticapm
from elasticapm.conf import constants
from elasticapm.contrib.starlette import ElasticAPM, make_apm_client

pytestmark = [pytest.mark.starlette]


@pytest.fixture
def fastapi_app(elasticapm_client):
    router = APIRouter(prefix="/api")

    @router.get("/items/{item_id}")
    async def get_item(item_id: int):
        return {"item_id": item_id}

    app = FastAPI()
    app.include_router(router)
    app.add_middleware(ElasticAPM, client=elasticapm_client)

    yield app

    elasticapm.uninstrument()


def test_included_router_request_succeeds(fastapi_app, elasticapm_client):
    client = TestClient(fastapi_app)

    response = client.get("/api/items/42")

    assert response.status_code == 200
    assert response.json() == {"item_id": 42}
    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1


def test_included_router_transaction_name(fastapi_app, elasticapm_client):
    client = TestClient(fastapi_app)

    response = client.get("/api/items/42")

    assert response.status_code == 200

    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    assert transaction["name"] == "GET /api/items/{item_id}"


def test_included_router_with_mounted_app_transaction_name(elasticapm_client):
    router = APIRouter(prefix="/api")

    @router.get("/items/{item_id}")
    async def get_item(item_id: int):
        return {"item_id": item_id}

    async def hi(request):
        return PlainTextResponse("sub")

    sub = Starlette(routes=[Route("/hi", hi)])
    app = FastAPI()
    app.include_router(router)
    app.mount("/sub", sub)
    app.add_middleware(ElasticAPM, client=elasticapm_client)
    client = TestClient(app)

    try:
        response = client.get("/sub/hi")

        assert response.status_code == 200
        assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
        transaction = elasticapm_client.events[constants.TRANSACTION][0]
        assert transaction["name"] == "GET /sub/hi"
    finally:
        elasticapm.uninstrument()


def test_included_router_options_partial_match(fastapi_app, elasticapm_client):
    client = TestClient(fastapi_app)

    response = client.options("/api/items/42")

    assert response.status_code == 405
    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1


def test_make_client_with_fastapi_framework():
    c = make_apm_client(config={"SERVICE_NAME": "foo"}, client_cls=TempStoreClient, framework_name="fastapi")
    c.close()
    assert c.config.service_name == "foo"



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest
from django.apps import apps

import elasticapm
from elasticapm.conf.constants import SPAN
from elasticapm.contrib.django.apps import instrument, register_handlers
from elasticapm.contrib.django.client import DjangoClient


class TempStoreClient(DjangoClient):
    def __init__(self, **inline) -> None:
        inline.setdefault("transport_class", "tests.fixtures.DummyTransport")
        super(TempStoreClient, self).__init__(**inline)

    @property
    def events(self):
        return self._transport.events

    def spans_for_transaction(self, transaction):
        """Test helper method to get all spans of a specific transaction"""
        return [span for span in self.events[SPAN] if span["transaction_id"] == transaction["id"]]


@pytest.fixture()
def django_elasticapm_client(request):
    client_config = getattr(request, "param", {})
    client_config.setdefault("service_name", "app")
    client_config.setdefault("secret_token", "secret")
    client_config.setdefault("span_stack_trace_min_duration", 0)
    client_config.setdefault("span_compression_exact_match_max_duration", "0ms")
    client_config.setdefault("span_compression_same_kind_max_duration", "0ms")
    app = apps.get_app_config("elasticapm")
    old_client = app.client
    client = TempStoreClient(**client_config)
    register_handlers(client)
    instrument(client)
    app.client = client
    yield client
    client.close()
    elasticapm.uninstrument()

    app.client = old_client

    if old_client:
        register_handlers(old_client)
        instrument(old_client)


@pytest.fixture()
def django_sending_elasticapm_client(request, validating_httpserver):
    validating_httpserver.serve_content(code=202, content="", headers={"Location": "http://example.com/foo"})
    client_config = getattr(request, "param", {})
    client_config.setdefault("server_url", validating_httpserver.url)
    client_config.setdefault("server_version", (8, 0, 0))
    client_config.setdefault("service_name", "app")
    client_config.setdefault("secret_token", "secret")
    client_config.setdefault("transport_class", "elasticapm.transport.http.Transport")
    client_config.setdefault("span_stack_trace_min_duration", 0)
    client_config.setdefault("span_compression_exact_match_max_duration", "0ms")
    client_config.setdefault("span_compression_same_kind_max_duration", "0ms")
    client_config.setdefault("exit_span_min_duration", "0ms")
    app = apps.get_app_config("elasticapm")
    old_client = app.client
    client = DjangoClient(**client_config)
    register_handlers(client)
    instrument(client)
    app.client = client
    client.httpserver = validating_httpserver
    yield client
    client.close()
    elasticapm.uninstrument()

    app.client = old_client

    if old_client:
        register_handlers(old_client)
        instrument(old_client)



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import

import logging
import time

from django.contrib.auth.models import User
from django.db import DatabaseError
from django.db.models import QuerySet
from django.http import HttpResponse, StreamingHttpResponse
from django.shortcuts import get_object_or_404, render
from django.views import View

import elasticapm


class MyException(Exception):
    pass


class IgnoredException(Exception):
    skip_elasticapm = True


def no_error(request, id=None):
    resp = HttpResponse(str(id))
    resp["My-Header"] = "foo"
    return resp


def http_error(request, status=None):
    if status:
        status = int(status)
    resp = HttpResponse(status=status)
    return resp


def fake_login(request):
    return HttpResponse("")


def django_exc(request):
    return get_object_or_404(MyException, pk=1)


def django_queryset_error(request):
    """Simulation of django ORM timeout"""

    class CustomQuerySet(QuerySet):
        def all(self):
            raise DatabaseError()

        def __repr__(self) -> str:
            return str(self._result_cache)

    qs = CustomQuerySet()
    list(qs.all())


def raise_exc(request):
    raise MyException(request.GET.get("message", "view exception"))


def raise_ioerror(request):
    raise IOError(request.GET.get("message", "view exception"))


def decorated_raise_exc(request):
    return raise_exc(request)


def template_exc(request):
    return render(request, "error.html")


def ignored_exception(request):
    raise IgnoredException()


def logging_request_exc(request):
    logger = logging.getLogger(__name__)
    try:
        raise Exception(request.GET.get("message", "view exception"))
    except Exception as e:
        logger.error(e, exc_info=True, extra={"request": request})
    return HttpResponse("")


def logging_view(request):
    logger = logging.getLogger("logmiddleware")
    logger.info("Just loggin'")
    return HttpResponse("")


def render_template_view(request):
    def something_expensive():
        with elasticapm.capture_span("something_expensive", "code"):
            return [User(username="Ron"), User(username="Beni")]

    return render(request, "list_users.html", {"users": something_expensive})


def render_jinja2_template(request):
    return render(request, "jinja2_template.html")


def render_user_view(request):
    def something_expensive():
        with elasticapm.capture_span("something_expensive", "code"):
            for i in range(100):
                users = list(User.objects.all())
        return users

    return render(request, "list_users.html", {"users": something_expensive})


def streaming_view(request):
    def my_generator():
        for i in range(5):
            with elasticapm.capture_span("iter", "code"):
                time.sleep(0.01)
                yield str(i)

    resp = StreamingHttpResponse(my_generator())
    return resp


def override_transaction_name_view(request):
    elasticapm.set_transaction_name("foo")
    elasticapm.set_transaction_result("okydoky")
    return HttpResponse()


class ClassBasedView(View):
    def get(self, request):
        return HttpResponse()



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import urllib.parse

from elasticapm.conf import constants
from elasticapm.instrumentation.packages.base import AbstractInstrumentedModule
from elasticapm.traces import DroppedSpan, capture_span, execution_context
from elasticapm.utils import default_ports, sanitize_url
from elasticapm.utils.disttracing import TracingOptions


# copied and adapted from urllib.request
def request_host(request):
    """Return request-host, as defined by RFC 2965.

    Variation from RFC: returned value is lowercased, for convenient
    comparison.

    """
    url = request.get_full_url()
    parse_result = urllib.parse.urlparse(url)
    scheme, host, port = parse_result.scheme, parse_result.hostname, parse_result.port
    try:
        port = int(port)
    except (ValueError, TypeError):
        pass
    if host == "":
        host = request.get_header("Host", "")

    if port and port != default_ports.get(scheme):
        host = "%s:%s" % (host, port)
    return host


class UrllibInstrumentation(AbstractInstrumentedModule):
    name = "urllib"

    instrument_list = [("urllib.request", "AbstractHTTPHandler.do_open")]

    def call(self, module, method, wrapped, instance, args, kwargs):
        request_object = args[1] if len(args) > 1 else kwargs["req"]

        method = request_object.get_method()
        host = request_host(request_object)

        url = sanitize_url(request_object.get_full_url())
        signature = method.upper() + " " + host

        transaction = execution_context.get_transaction()

        with capture_span(
            signature,
            span_type="external",
            span_subtype="http",
            extra={"http": {"url": url}},
            leaf=True,
        ) as span:
            # if urllib has been called in a leaf span, this span might be a DroppedSpan.
            leaf_span = span
            while isinstance(leaf_span, DroppedSpan):
                leaf_span = leaf_span.parent

            parent_id = leaf_span.id if leaf_span else transaction.id
            trace_parent = transaction.trace_parent.copy_from(
                span_id=parent_id, trace_options=TracingOptions(recorded=True)
            )
            self._set_disttracing_headers(request_object, trace_parent, transaction)
            if leaf_span:
                leaf_span.dist_tracing_propagated = True
            response = wrapped(*args, **kwargs)
            if response:
                if span.context:
                    span.context["http"]["status_code"] = response.status
                span.set_success() if response.status < 400 else span.set_failure()
            return response

    def mutate_unsampled_call_args(self, module, method, wrapped, instance, args, kwargs, transaction):
        request_object = args[1] if len(args) > 1 else kwargs["req"]
        # since we don't have a span, we set the span id to the transaction id
        trace_parent = transaction.trace_parent.copy_from(
            span_id=transaction.id, trace_options=TracingOptions(recorded=False)
        )

        self._set_disttracing_headers(request_object, trace_parent, transaction)
        return args, kwargs

    def _set_disttracing_headers(self, request_object, trace_parent, transaction) -> None:
        trace_parent_str = trace_parent.to_string()
        request_object.add_header(constants.TRACEPARENT_HEADER_NAME, trace_parent_str)
        if transaction.tracer.config.use_elastic_traceparent_header:
            request_object.add_header(constants.TRACEPARENT_LEGACY_HEADER_NAME, trace_parent_str)
        if trace_parent.tracestate:
            request_object.add_header(constants.TRACESTATE_HEADER_NAME, trace_parent.tracestate)



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os

import pytest

from elasticapm.metrics.base_metrics import MetricsRegistry

try:
    from elasticapm.metrics.sets.cpu_linux import CPUMetricSet
except ImportError:
    pytest.skip("Not a Linux system", allow_module_level=True)


TEMPLATE_PROC_STAT_SELF = """32677 (python) R 5333 32677 5333 34822 32677 4194304 13815 176 2 0 {utime} {stime} 0 0 20 0 7 0 6010710 3686981632 11655 18446744073709551615 94580847771648 94580849947501 140732830512176 0 0 0 0 16781312 134217730 0 0 0 17 1 0 0 6 0 0 94580850578256 94580851016824 94580875862016 140732830518932 140732830518950 140732830518950 140732830523339 0"""


TEMPLATE_PROC_STAT_DEBIAN = """cpu  {user} 2037 278561 {idle} 15536 0 178811 0 0 0
cpu0 189150 166 34982 1081369 2172 0 73586 0 0 0
cpu1 190286 201 35110 637359 1790 0 41941 0 0 0
intr 60591079 11 12496 0 0 0 0 0 0 1 27128 0 0 474144 0 0 0 0 0 0 0 0 0 0
ctxt 215687788
btime 1544981001
processes 416902
procs_running 3
procs_blocked 1
softirq 61270136 2508862 17040144 383 7315609 6627 56 23515 19184746 49637 15140557
"""


TEMPLATE_PROC_STAT_RHEL = """cpu  {user} 2037 278561 {idle} 15536 0 178811
cpu0 259246 7001 60190 34250993 137517 772 0
intr 354133732 347209999 2272 0 4 4 0 0 3 1 1249247 0 0 80143 0 422626 5169433
ctxt 12547729
btime 1093631447
processes 130523
procs_running 1
procs_blocked 0"""


TEMPLATE_PROC_MEMINFO = """MemTotal:       16164884 kB
MemFree:          359184 kB
MemAvailable:    6514428 kB
Buffers:          891296 kB
Cached:          6416340 kB
SwapCached:          148 kB
Active:          7682276 kB
Inactive:        5468500 kB
"""


TEMPLATE_PROC_MEMINFO_NO_MEMAVAILABLE = """MemTotal:       16164884 kB
MemFree:          359184 kB
Buffers:          891296 kB
Cached:          6416340 kB
SwapCached:          148 kB
Active:          7682276 kB
Inactive:        5468500 kB
"""

TEMPLATE_CGROUP_MEM_LIMIT_IN_BYTES = "9223372036854771712"
TEMPLATE_CGROUP_MEM_LIMIT_IN_BYTES_LIMITED = "7964778496"
TEMPLATE_CGROUP_MEM_USAGE_IN_BYTES = "964778496"

TEMPLATE_CGROUP_MEM_STAT = """pgmajfault 0
inactive_anon 0
active_anon 778702848
inactive_file 10407936
active_file 0
unevictable 0
hierarchical_memory_limit 1073741824
hierarchical_memsw_limit 2147483648
"""


@pytest.mark.parametrize("proc_stat_template", [TEMPLATE_PROC_STAT_DEBIAN, TEMPLATE_PROC_STAT_RHEL])
def test_cpu_mem_from_proc(elasticapm_client, proc_stat_template, tmpdir):
    proc_stat_self = os.path.join(tmpdir.strpath, "self-stat")
    proc_stat = os.path.join(tmpdir.strpath, "stat")
    proc_meminfo = os.path.join(tmpdir.strpath, "meminfo")
    cgroup_memory_limit = os.path.join(tmpdir.strpath, "memory", "memory.limit_in_bytes")
    cgroup_memory_usage = os.path.join(tmpdir.strpath, "memory", "memory.usage_in_bytes")
    proc_self_cgroup = os.path.join(tmpdir.strpath, "cgroup")
    os.mkdir(os.path.join(tmpdir.strpath, "memory"))
    proc_self_mount = os.path.join(tmpdir.strpath, "mountinfo")

    for path, content in (
        (proc_stat, proc_stat_template.format(user=0, idle=0)),
        (proc_stat_self, TEMPLATE_PROC_STAT_SELF.format(utime=0, stime=0)),
        (proc_meminfo, TEMPLATE_PROC_MEMINFO),
        (cgroup_memory_limit, TEMPLATE_CGROUP_MEM_LIMIT_IN_BYTES),
        (cgroup_memory_usage, TEMPLATE_CGROUP_MEM_USAGE_IN_BYTES),
        (proc_self_cgroup, "9:memory:/slice"),
        (
            proc_self_mount,
            "39 30 0:35 / "
            + tmpdir.strpath
            + "/memory rw,nosuid,nodev,noexec,relatime shared:10 - cgroup cgroup rw,seclabel,memory\n",
        ),
    ):
        with open(path, mode="w") as f:
            f.write(content)
    metricset = CPUMetricSet(
        MetricsRegistry(elasticapm_client),
        sys_stats_file=proc_stat,
        process_stats_file=proc_stat_self,
        memory_stats_file=proc_meminfo,
        proc_self_cgroup=proc_self_cgroup,
        mount_info=proc_self_mount,
    )

    for path, content in (
        (proc_stat, proc_stat_template.format(user=400000, idle=600000)),
        (proc_stat_self, TEMPLATE_PROC_STAT_SELF.format(utime=100000, stime=100000)),
        (proc_meminfo, TEMPLATE_PROC_MEMINFO),
    ):
        with open(path, mode="w") as f:
            f.write(content)
    data = next(metricset.collect())
    assert data["samples"]["system.cpu.total.norm.pct"]["value"] == 0.4
    assert data["samples"]["system.process.cpu.total.norm.pct"]["value"] == 0.2

    assert data["samples"]["system.memory.total"]["value"] == 16552841216
    assert data["samples"]["system.memory.actual.free"]["value"] == 6670774272

    assert data["samples"]["system.process.memory.rss.bytes"]["value"] == 47738880
    assert data["samples"]["system.process.memory.size"]["value"] == 3686981632


def test_mem_free_from_memfree_when_memavailable_not_mentioned(elasticapm_client, tmpdir):
    proc_stat_self = os.path.join(tmpdir.strpath, "self-stat")
    proc_stat = os.path.join(tmpdir.strpath, "stat")
    proc_meminfo = os.path.join(tmpdir.strpath, "meminfo")

    for path, content in (
        (proc_stat, TEMPLATE_PROC_STAT_DEBIAN.format(user=0, idle=0)),
        (proc_stat_self, TEMPLATE_PROC_STAT_SELF.format(utime=0, stime=0)),
        (proc_meminfo, TEMPLATE_PROC_MEMINFO_NO_MEMAVAILABLE),
    ):
        with open(path, mode="w") as f:
            f.write(content)
    metricset = CPUMetricSet(
        MetricsRegistry(elasticapm_client),
        sys_stats_file=proc_stat,
        process_stats_file=proc_stat_self,
        memory_stats_file=proc_meminfo,
    )

    for path, content in (
        (proc_stat, TEMPLATE_PROC_STAT_DEBIAN.format(user=400000, idle=600000)),
        (proc_stat_self, TEMPLATE_PROC_STAT_SELF.format(utime=100000, stime=100000)),
        (proc_meminfo, TEMPLATE_PROC_MEMINFO_NO_MEMAVAILABLE),
    ):
        with open(path, mode="w") as f:
            f.write(content)
    data = next(metricset.collect())

    mem_free_expected = sum([359184, 891296, 6416340]) * 1024  # MemFree + Buffers + Cached, in bytes
    assert data["samples"]["system.memory.actual.free"]["value"] == mem_free_expected


def test_cpu_usage_when_cpu_total_is_zero(elasticapm_client, tmpdir):
    proc_stat_self = os.path.join(tmpdir.strpath, "self-stat")
    proc_stat = os.path.join(tmpdir.strpath, "stat")
    proc_meminfo = os.path.join(tmpdir.strpath, "meminfo")

    for path, content in (
        (proc_stat, TEMPLATE_PROC_STAT_DEBIAN.format(user=0, idle=0)),
        (proc_stat_self, TEMPLATE_PROC_STAT_SELF.format(utime=0, stime=0)),
        (proc_meminfo, TEMPLATE_PROC_MEMINFO_NO_MEMAVAILABLE),
    ):
        with open(path, mode="w") as f:
            f.write(content)
    metricset = CPUMetricSet(
        MetricsRegistry(elasticapm_client),
        sys_stats_file=proc_stat,
        process_stats_file=proc_stat_self,
        memory_stats_file=proc_meminfo,
    )
    data = next(metricset.collect())

    cpu_total_expected = 0
    assert data["samples"]["system.cpu.total.norm.pct"]["value"] == cpu_total_expected
    assert data["samples"]["system.process.cpu.total.norm.pct"]["value"] == cpu_total_expected


def test_mem_from_cgroup1(elasticapm_client, tmpdir):
    proc_stat_self = os.path.join(tmpdir.strpath, "self-stat")
    proc_stat = os.path.join(tmpdir.strpath, "stat")
    proc_meminfo = os.path.join(tmpdir.strpath, "meminfo")
    cgroup_memory_limit = os.path.join(tmpdir.strpath, "memory", "memory.limit_in_bytes")
    cgroup_memory_usage = os.path.join(tmpdir.strpath, "memory", "memory.usage_in_bytes")
    cgroup_memory_stat = os.path.join(tmpdir.strpath, "memory", "memory.stat")
    proc_self_cgroup = os.path.join(tmpdir.strpath, "cgroup")
    os.mkdir(os.path.join(tmpdir.strpath, "memory"))
    proc_self_mount = os.path.join(tmpdir.strpath, "mountinfo")

    for path, content in (
        (proc_stat, TEMPLATE_PROC_STAT_DEBIAN.format(user=0, idle=0)),
        (proc_stat_self, TEMPLATE_PROC_STAT_SELF.format(utime=0, stime=0)),
        (proc_meminfo, TEMPLATE_PROC_MEMINFO),
        (cgroup_memory_limit, TEMPLATE_CGROUP_MEM_LIMIT_IN_BYTES_LIMITED),
        (cgroup_memory_usage, TEMPLATE_CGROUP_MEM_USAGE_IN_BYTES),
        (cgroup_memory_stat, TEMPLATE_CGROUP_MEM_STAT),
        (proc_self_cgroup, "9:memory:/slice"),
        (
            proc_self_mount,
            "39 30 0:35 / "
            + tmpdir.strpath
            + "/memory rw,nosuid,nodev,noexec,relatime shared:10 - cgroup cgroup rw,seclabel,memory\n",
        ),
    ):
        with open(path, mode="w") as f:
            f.write(content)
    metricset = CPUMetricSet(
        MetricsRegistry(elasticapm_client),
        sys_stats_file=proc_stat,
        process_stats_file=proc_stat_self,
        memory_stats_file=proc_meminfo,
        proc_self_cgroup=proc_self_cgroup,
        mount_info=proc_self_mount,
    )

    data = next(metricset.collect())

    assert data["samples"]["system.memory.total"]["value"] == 16552841216
    assert data["samples"]["system.memory.actual.free"]["value"] == 6670774272

    assert data["samples"]["system.process.memory.rss.bytes"]["value"] == 47738880
    assert data["samples"]["system.process.memory.size"]["value"] == 3686981632

    assert data["samples"]["system.process.cgroup.memory.mem.limit.bytes"]["value"] == 7964778496
    assert data["samples"]["system.process.cgroup.memory.mem.usage.bytes"]["value"] == 964778496


def test_mem_from_cgroup2(elasticapm_client, tmpdir):
    proc_stat_self = os.path.join(tmpdir.strpath, "self-stat")
    proc_stat = os.path.join(tmpdir.strpath, "stat")
    proc_meminfo = os.path.join(tmpdir.strpath, "meminfo")
    cgroup2_memory_limit = os.path.join(tmpdir.strpath, "slice", "memory.max")
    cgroup2_memory_usage = os.path.join(tmpdir.strpath, "slice", "memory.current")
    cgroup2_memory_stat = os.path.join(tmpdir.strpath, "slice", "memory.stat")
    cgroup2_self_cgroup = os.path.join(tmpdir.strpath, "cgroup")
    proc_self_cgroup = os.path.join(tmpdir.strpath, "cgroup")
    os.mkdir(os.path.join(tmpdir.strpath, "slice"))
    proc_self_mount = os.path.join(tmpdir.strpath, "mountinfo")

    for path, content in (
        (proc_stat, TEMPLATE_PROC_STAT_DEBIAN.format(user=0, idle=0)),
        (proc_stat_self, TEMPLATE_PROC_STAT_SELF.format(utime=0, stime=0)),
        (proc_meminfo, TEMPLATE_PROC_MEMINFO),
        (cgroup2_memory_limit, TEMPLATE_CGROUP_MEM_LIMIT_IN_BYTES_LIMITED),
        (cgroup2_memory_usage, TEMPLATE_CGROUP_MEM_USAGE_IN_BYTES),
        (cgroup2_memory_stat, TEMPLATE_CGROUP_MEM_STAT),
        (cgroup2_self_cgroup, "9:memory:/slice"),
        (
            proc_self_mount,
            "30 23 0:26 / "
            + tmpdir.strpath
            + " rw,nosuid,nodev,noexec,relatime shared:4 - cgroup2 cgroup rw,seclabel\n",
        ),
    ):
        with open(path, mode="w") as f:
            f.write(content)
    metricset = CPUMetricSet(
        MetricsRegistry(elasticapm_client),
        sys_stats_file=proc_stat,
        process_stats_file=proc_stat_self,
        memory_stats_file=proc_meminfo,
        proc_self_cgroup=proc_self_cgroup,
        mount_info=proc_self_mount,
    )

    data = next(metricset.collect())

    assert data["samples"]["system.memory.total"]["value"] == 16552841216
    assert data["samples"]["system.memory.actual.free"]["value"] == 6670774272

    assert data["samples"]["system.process.memory.rss.bytes"]["value"] == 47738880
    assert data["samples"]["system.process.memory.size"]["value"] == 3686981632

    assert data["samples"]["system.process.cgroup.memory.mem.limit.bytes"]["value"] == 7964778496
    assert data["samples"]["system.process.cgroup.memory.mem.usage.bytes"]["value"] == 964778496


def test_mem_from_cgroup1_max_handling(elasticapm_client, tmpdir):
    proc_stat_self = os.path.join(tmpdir.strpath, "self-stat")
    proc_stat = os.path.join(tmpdir.strpath, "stat")
    proc_meminfo = os.path.join(tmpdir.strpath, "meminfo")
    cgroup_memory_limit = os.path.join(tmpdir.strpath, "memory", "memory.limit_in_bytes")
    cgroup_memory_usage = os.path.join(tmpdir.strpath, "memory", "memory.usage_in_bytes")
    cgroup_memory_stat = os.path.join(tmpdir.strpath, "memory", "memory.stat")
    proc_self_cgroup = os.path.join(tmpdir.strpath, "cgroup")
    os.mkdir(os.path.join(tmpdir.strpath, "memory"))
    proc_self_mount = os.path.join(tmpdir.strpath, "mountinfo")

    for path, content in (
        (proc_stat, TEMPLATE_PROC_STAT_DEBIAN.format(user=0, idle=0)),
        (proc_stat_self, TEMPLATE_PROC_STAT_SELF.format(utime=0, stime=0)),
        (proc_meminfo, TEMPLATE_PROC_MEMINFO),
        (cgroup_memory_limit, str(0x7FFFFFFFFFFFF000)),
        (cgroup_memory_usage, TEMPLATE_CGROUP_MEM_USAGE_IN_BYTES),
        (cgroup_memory_stat, TEMPLATE_CGROUP_MEM_STAT),
        (proc_self_cgroup, "9:memory:/slice"),
        (
            proc_self_mount,
            "39 30 0:35 / "
            + tmpdir.strpath
            + "/memory rw,nosuid,nodev,noexec,relatime shared:10 - cgroup cgroup rw,seclabel,memory\n",
        ),
    ):
        with open(path, mode="w") as f:
            f.write(content)
    metricset = CPUMetricSet(
        MetricsRegistry(elasticapm_client),
        sys_stats_file=proc_stat,
        process_stats_file=proc_stat_self,
        memory_stats_file=proc_meminfo,
        proc_self_cgroup=proc_self_cgroup,
        mount_info=proc_self_mount,
    )

    data = next(metricset.collect())

    assert "system.process.cgroup.memory.mem.limit.bytes" not in data["samples"]
    assert "system.process.cgroup.memory.mem.usage.bytes" not in data["samples"]


def test_mem_from_cgroup2_max_handling(elasticapm_client, tmpdir):
    proc_stat_self = os.path.join(tmpdir.strpath, "self-stat")
    proc_stat = os.path.join(tmpdir.strpath, "stat")
    proc_meminfo = os.path.join(tmpdir.strpath, "meminfo")
    cgroup2_memory_limit = os.path.join(tmpdir.strpath, "slice", "memory.max")
    cgroup2_memory_usage = os.path.join(tmpdir.strpath, "slice", "memory.current")
    cgroup2_memory_stat = os.path.join(tmpdir.strpath, "slice", "memory.stat")
    cgroup2_self_cgroup = os.path.join(tmpdir.strpath, "cgroup")
    proc_self_cgroup = os.path.join(tmpdir.strpath, "cgroup")
    os.mkdir(os.path.join(tmpdir.strpath, "slice"))
    proc_self_mount = os.path.join(tmpdir.strpath, "mountinfo")

    for path, content in (
        (proc_stat, TEMPLATE_PROC_STAT_DEBIAN.format(user=0, idle=0)),
        (proc_stat_self, TEMPLATE_PROC_STAT_SELF.format(utime=0, stime=0)),
        (proc_meminfo, TEMPLATE_PROC_MEMINFO),
        (cgroup2_memory_limit, "max"),
        (cgroup2_memory_usage, TEMPLATE_CGROUP_MEM_USAGE_IN_BYTES),
        (cgroup2_memory_stat, TEMPLATE_CGROUP_MEM_STAT),
        (cgroup2_self_cgroup, "9:memory:/slice"),
        (
            proc_self_mount,
            "30 23 0:26 / "
            + tmpdir.strpath
            + " rw,nosuid,nodev,noexec,relatime shared:4 - cgroup2 cgroup rw,seclabel\n",
        ),
    ):
        with open(path, mode="w") as f:
            f.write(content)
    metricset = CPUMetricSet(
        MetricsRegistry(elasticapm_client),
        sys_stats_file=proc_stat,
        process_stats_file=proc_stat_self,
        memory_stats_file=proc_meminfo,
        proc_self_cgroup=proc_self_cgroup,
        mount_info=proc_self_mount,
    )

    data = next(metricset.collect())

    assert "system.process.cgroup.memory.mem.limit.bytes" not in data["samples"]
    assert "system.process.cgroup.memory.mem.usage.bytes" not in data["samples"]


def test_mem_from_cgroup_files_dont_exist(elasticapm_client, tmpdir):
    # it appears that on Google App engine, there is a possibility of
    # memory.limit_in_bytes existing while memory.usage_in_bytes does not.
    # See https://github.com/elastic/apm-agent-python/issues/985
    proc_stat_self = os.path.join(tmpdir.strpath, "self-stat")
    proc_stat = os.path.join(tmpdir.strpath, "stat")
    proc_meminfo = os.path.join(tmpdir.strpath, "meminfo")
    cgroup_memory_limit = os.path.join(tmpdir.strpath, "memory", "memory.limit_in_bytes")
    # intentionally commented out
    # cgroup_memory_usage = os.path.join(tmpdir.strpath, "memory", "memory.usage_in_bytes")
    cgroup_memory_stat = os.path.join(tmpdir.strpath, "memory", "memory.stat")
    proc_self_cgroup = os.path.join(tmpdir.strpath, "cgroup")
    os.mkdir(os.path.join(tmpdir.strpath, "memory"))
    proc_self_mount = os.path.join(tmpdir.strpath, "mountinfo")

    for path, content in (
        (proc_stat, TEMPLATE_PROC_STAT_DEBIAN.format(user=0, idle=0)),
        (proc_stat_self, TEMPLATE_PROC_STAT_SELF.format(utime=0, stime=0)),
        (proc_meminfo, TEMPLATE_PROC_MEMINFO),
        (cgroup_memory_limit, TEMPLATE_CGROUP_MEM_LIMIT_IN_BYTES_LIMITED),
        # intentionally commented out
        # (cgroup_memory_usage, TEMPLATE_CGROUP_MEM_USAGE_IN_BYTES),
        (cgroup_memory_stat, TEMPLATE_CGROUP_MEM_STAT),
        (proc_self_cgroup, "9:memory:/slice"),
        (
            proc_self_mount,
            "39 30 0:35 / "
            + tmpdir.strpath
            + "/memory rw,nosuid,nodev,noexec,relatime shared:10 - cgroup cgroup rw,seclabel,memory\n",
        ),
    ):
        with open(path, mode="w") as f:
            f.write(content)
    metricset = CPUMetricSet(
        MetricsRegistry(elasticapm_client),
        sys_stats_file=proc_stat,
        process_stats_file=proc_stat_self,
        memory_stats_file=proc_meminfo,
        proc_self_cgroup=proc_self_cgroup,
        mount_info=proc_self_mount,
    )

    assert metricset.cgroup_files.limit is not None
    assert metricset.cgroup_files.usage is None

    data = next(metricset.collect())

    assert "system.process.cgroup.memory.mem.limit.bytes" in data["samples"]
    assert "system.process.cgroup.memory.mem.usage.bytes" not in data["samples"]



#  BSD 3-Clause License
#
#  Copyright (c) 2022, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import time
from collections import defaultdict

import pytest

import elasticapm
from elasticapm.conf import constants
from elasticapm.utils import encoding
from elasticapm.utils.disttracing import TraceParent


@pytest.mark.parametrize(
    "elasticapm_client", [{"server_version": (7, 15, 0)}, {"server_version": (7, 16, 0)}], indirect=True
)
def test_transaction_span(elasticapm_client):
    elasticapm_client.begin_transaction("test")
    with elasticapm.capture_span("test", extra={"a": "b"}):
        pass
    elasticapm_client.end_transaction("test", constants.OUTCOME.SUCCESS)
    transactions = elasticapm_client.events[constants.TRANSACTION]
    assert len(transactions) == 1
    assert transactions[0]["name"] == "test"
    assert transactions[0]["type"] == "test"
    assert transactions[0]["result"] == constants.OUTCOME.SUCCESS

    spans = elasticapm_client.events[constants.SPAN]
    assert len(spans) == 1
    assert spans[0]["name"] == "test"


@pytest.mark.parametrize(
    "elasticapm_client", [{"transactions_ignore_patterns": ["^OPTIONS", "views.api.v2"]}], indirect=True
)
def test_ignore_patterns(elasticapm_client):
    elasticapm_client.begin_transaction("web")
    elasticapm_client.end_transaction("OPTIONS views.healthcheck", 200)

    elasticapm_client.begin_transaction("web")
    elasticapm_client.end_transaction("GET views.users", 200)

    transactions = elasticapm_client.events[constants.TRANSACTION]

    assert len(transactions) == 1
    assert transactions[0]["name"] == "GET views.users"


@pytest.mark.parametrize(
    "elasticapm_client", [{"transactions_ignore_patterns": ["^OPTIONS", "views.api.v2"]}], indirect=True
)
def test_ignore_patterns_with_none_transaction_name(elasticapm_client):
    elasticapm_client.begin_transaction("web")
    t = elasticapm_client.end_transaction(None, 200)
    assert t.name == ""


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        {"collect_local_variables": "errors"},
        {"collect_local_variables": "transactions", "local_var_max_length": 20, "local_var_max_list_length": 10},
        {"collect_local_variables": "all", "local_var_max_length": 20, "local_var_max_list_length": 10},
        {"collect_local_variables": "something"},
    ],
    indirect=True,
)
def test_collect_local_variables_transactions(elasticapm_client):
    mode = elasticapm_client.config.collect_local_variables
    elasticapm_client.begin_transaction("test")
    with elasticapm.capture_span("foo"):
        a_local_var = 1
        a_long_local_var = 100 * "a"
        a_long_local_list = list(range(100))
        pass
    elasticapm_client.end_transaction("test", "ok")
    frame = elasticapm_client.events[constants.SPAN][0]["stacktrace"][0]
    if mode in ("transactions", "all"):
        assert "vars" in frame, mode
        assert frame["vars"]["a_local_var"] == 1
        assert len(frame["vars"]["a_long_local_var"]) == 20
        assert len(frame["vars"]["a_long_local_list"]) == 12
        assert frame["vars"]["a_long_local_list"][-1] == "(90 more elements)"
    else:
        assert "vars" not in frame, mode


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        {"source_lines_span_library_frames": 0, "source_lines_span_app_frames": 0},
        {"source_lines_span_library_frames": 1, "source_lines_span_app_frames": 1},
        {"source_lines_span_library_frames": 7, "source_lines_span_app_frames": 5},
    ],
    indirect=True,
)
def test_collect_source_transactions(elasticapm_client):
    library_frame_context = elasticapm_client.config.source_lines_span_library_frames
    in_app_frame_context = elasticapm_client.config.source_lines_span_app_frames
    elasticapm_client.begin_transaction("test")
    with elasticapm.capture_span("foo"):
        pass
    elasticapm_client.end_transaction("test", "ok")
    span = elasticapm_client.events[constants.SPAN][0]
    in_app_frame = span["stacktrace"][0]
    library_frame = span["stacktrace"][1]
    assert not in_app_frame["library_frame"]
    assert library_frame["library_frame"]
    if library_frame_context:
        assert "context_line" in library_frame, library_frame_context
        assert "pre_context" in library_frame, library_frame_context
        assert "post_context" in library_frame, library_frame_context
        lines = len([library_frame["context_line"]] + library_frame["pre_context"] + library_frame["post_context"])
        assert lines == library_frame_context, library_frame_context
    else:
        assert "context_line" not in library_frame, library_frame_context
        assert "pre_context" not in library_frame, library_frame_context
        assert "post_context" not in library_frame, library_frame_context
    if in_app_frame_context:
        assert "context_line" in in_app_frame, in_app_frame_context
        assert "pre_context" in in_app_frame, in_app_frame_context
        assert "post_context" in in_app_frame, in_app_frame_context
        lines = len([in_app_frame["context_line"]] + in_app_frame["pre_context"] + in_app_frame["post_context"])
        assert lines == in_app_frame_context, (in_app_frame_context, in_app_frame["lineno"])
    else:
        assert "context_line" not in in_app_frame, in_app_frame_context
        assert "pre_context" not in in_app_frame, in_app_frame_context
        assert "post_context" not in in_app_frame, in_app_frame_context


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        {"transaction_sample_rate": 0.4, "server_version": (7, 14, 0)},
        {"transaction_sample_rate": 0.4, "server_version": (8, 0, 0)},
    ],
    indirect=True,
)
def test_transaction_sampling(elasticapm_client, not_so_random):
    for i in range(10):
        elasticapm_client.begin_transaction("test_type")
        with elasticapm.capture_span("xyz"):
            pass
        elasticapm_client.end_transaction("test")

    transactions = elasticapm_client.events[constants.TRANSACTION]
    spans_per_transaction = defaultdict(list)
    for span in elasticapm_client.events[constants.SPAN]:
        spans_per_transaction[span["transaction_id"]].append(span)

    # seed is fixed by not_so_random fixture
    assert len([t for t in transactions if t["sampled"]]) == 3
    if elasticapm_client.server_version < (8, 0, 0):
        assert len(transactions) == 10
    else:
        assert len(transactions) == 3
    for transaction in transactions:
        assert transaction["sampled"] or not transaction["id"] in spans_per_transaction
        assert transaction["sampled"] or not "context" in transaction
        assert transaction["sample_rate"] == 0 if not transaction["sampled"] else transaction["sample_rate"] == 0.4


def test_transaction_sample_rate_dynamic(elasticapm_client, not_so_random):
    elasticapm_client.config.update(version="1", transaction_sample_rate=0.4)
    for i in range(10):
        elasticapm_client.begin_transaction("test_type")
        with elasticapm.capture_span("xyz"):
            pass
        elasticapm_client.end_transaction("test")

    transactions = elasticapm_client.events[constants.TRANSACTION]
    spans_per_transaction = defaultdict(list)
    for span in elasticapm_client.events[constants.SPAN]:
        spans_per_transaction[span["transaction_id"]].append(span)

    # seed is fixed by not_so_random fixture
    assert len([t for t in transactions if t["sampled"]]) == 3
    for transaction in transactions:
        assert transaction["sampled"] or not transaction["id"] in spans_per_transaction
        assert transaction["sampled"] or not "context" in transaction

    elasticapm_client.config.update(version="1", transaction_sample_rate=1.0)
    for i in range(5):
        elasticapm_client.begin_transaction("test_type")
        with elasticapm.capture_span("xyz"):
            pass
        elasticapm_client.end_transaction("test")

    transactions = elasticapm_client.events[constants.TRANSACTION]

    # seed is fixed by not_so_random fixture
    assert len([t for t in transactions if t["sampled"]]) == 8


def test_transaction_max_spans_dynamic(elasticapm_client):
    elasticapm_client.config.update(version=1, transaction_max_spans=1)
    elasticapm_client.begin_transaction("test_type")
    for i in range(5):
        with elasticapm.capture_span("span"):
            pass
    elasticapm_client.end_transaction("test")
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)
    assert len(spans) == 1

    elasticapm_client.config.update(version=2, transaction_max_spans=3)
    elasticapm_client.begin_transaction("test_type")
    for i in range(5):
        with elasticapm.capture_span("span"):
            pass

    elasticapm_client.end_transaction("test")
    transaction = elasticapm_client.events[constants.TRANSACTION][1]
    spans = elasticapm_client.spans_for_transaction(transaction)
    assert len(spans) == 3


@pytest.mark.parametrize("elasticapm_client", [{"span_stack_trace_min_duration": 20}], indirect=True)
def test_transaction_span_stack_trace_min_duration(elasticapm_client):
    elasticapm_client.begin_transaction("test_type")
    with elasticapm.capture_span("noframes", duration=0.001):
        pass
    with elasticapm.capture_span("frames", duration=0.04):
        pass
    elasticapm_client.end_transaction("test")

    spans = elasticapm_client.events[constants.SPAN]

    assert len(spans) == 2
    assert spans[0]["name"] == "noframes"
    assert "stacktrace" not in spans[0]

    assert spans[1]["name"] == "frames"
    assert spans[1]["stacktrace"] is not None


@pytest.mark.parametrize("elasticapm_client", [{"span_stack_trace_min_duration": 0}], indirect=True)
def test_transaction_span_stack_trace_min_duration_no_limit(elasticapm_client):
    elasticapm_client.begin_transaction("test_type")
    with elasticapm.capture_span("frames"):
        pass
    with elasticapm.capture_span("frames", duration=0.04):
        pass
    elasticapm_client.end_transaction("test")

    spans = elasticapm_client.events[constants.SPAN]

    assert len(spans) == 2
    assert spans[0]["name"] == "frames"
    assert spans[0]["stacktrace"] is not None

    assert spans[1]["name"] == "frames"
    assert spans[1]["stacktrace"] is not None


def test_transaction_span_stack_trace_min_duration_dynamic(elasticapm_client):
    elasticapm_client.config.update(version="1", span_stack_trace_min_duration=20)
    elasticapm_client.begin_transaction("test_type")
    with elasticapm.capture_span("noframes", duration=0.001):
        pass
    with elasticapm.capture_span("frames", duration=0.04):
        pass
    elasticapm_client.end_transaction("test")

    spans = elasticapm_client.events[constants.SPAN]

    assert len(spans) == 2
    assert spans[0]["name"] == "noframes"
    assert "stacktrace" not in spans[0]

    assert spans[1]["name"] == "frames"
    assert spans[1]["stacktrace"] is not None

    elasticapm_client.config.update(version="1", span_stack_trace_min_duration=0)
    elasticapm_client.begin_transaction("test_type")
    with elasticapm.capture_span("frames"):
        pass
    with elasticapm.capture_span("frames", duration=0.04):
        pass
    elasticapm_client.end_transaction("test")

    spans = elasticapm_client.events[constants.SPAN]

    assert len(spans) == 4
    assert spans[2]["name"] == "frames"
    assert spans[2]["stacktrace"] is not None

    assert spans[3]["name"] == "frames"
    assert spans[3]["stacktrace"] is not None


def test_transaction_span_stack_trace_min_duration_overrides_old_config(elasticapm_client):
    """
    span_stack_trace_min_duration overrides span_frames_min_duration (which is deprecated)
    """
    elasticapm_client.config.update(version="1", span_stack_trace_min_duration=20, span_frames_min_duration=1)
    elasticapm_client.begin_transaction("test_type")
    with elasticapm.capture_span("noframes", duration=0.01):
        pass
    with elasticapm.capture_span("frames", duration=0.04):
        pass
    elasticapm_client.end_transaction("test")

    spans = elasticapm_client.events[constants.SPAN]

    assert len(spans) == 2
    assert spans[0]["name"] == "noframes"
    assert "stacktrace" not in spans[0]

    assert spans[1]["name"] == "frames"
    assert spans[1]["stacktrace"] is not None

    # Set span_stack_trace_min_duration to default so it picks up the non-default span_frames_min_duration
    elasticapm_client.config.update(version="1", span_stack_trace_min_duration=5, span_frames_min_duration=1)
    elasticapm_client.begin_transaction("test_type")
    with elasticapm.capture_span("yesframes", duration=0.01):
        pass
    with elasticapm.capture_span("frames", duration=0.04):
        pass
    elasticapm_client.end_transaction("test")

    spans = elasticapm_client.events[constants.SPAN]

    assert len(spans) == 4
    assert spans[2]["name"] == "yesframes"
    assert spans[2]["stacktrace"] is not None

    assert spans[3]["name"] == "frames"
    assert spans[3]["stacktrace"] is not None


def test_transaction_keyword_truncation(elasticapm_client):
    too_long = "x" * (constants.KEYWORD_MAX_LENGTH + 1)
    expected = encoding.keyword_field(too_long)
    assert too_long != expected
    assert len(expected) == constants.KEYWORD_MAX_LENGTH
    assert expected[-1] != "x"
    elasticapm_client.begin_transaction(too_long)
    elasticapm.label(val=too_long)
    elasticapm.set_user_context(username=too_long, email=too_long, user_id=too_long)
    with elasticapm.capture_span(name=too_long, span_type=too_long):
        pass
    elasticapm_client.end_transaction(too_long, too_long)
    elasticapm_client.close()

    span = elasticapm_client.events["span"][0]
    transaction = elasticapm_client.events["transaction"][0]

    assert transaction["name"] == expected
    assert transaction["type"] == expected
    assert transaction["result"] == expected

    assert transaction["context"]["user"]["id"] == expected
    assert transaction["context"]["user"]["username"] == expected
    assert transaction["context"]["user"]["email"] == expected

    assert transaction["context"]["tags"]["val"] == expected

    assert span["type"] == expected
    assert span["name"] == expected


def test_trace_parent(elasticapm_client):
    trace_parent = TraceParent.from_string("00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03")
    elasticapm_client.begin_transaction("test", trace_parent=trace_parent)
    transaction = elasticapm_client.end_transaction("test", "OK")
    data = transaction.to_dict()
    assert data["trace_id"] == "0af7651916cd43dd8448eb211c80319c"
    assert data["parent_id"] == "b7ad6b7169203331"


def test_sample_rate_in_dict(elasticapm_client):
    trace_parent = TraceParent.from_string(
        "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03", tracestate_string="es=s:0.43"
    )
    elasticapm_client.begin_transaction("test", trace_parent=trace_parent)
    with elasticapm.capture_span("x"):
        pass
    transaction = elasticapm_client.end_transaction("test", "OK")
    data = transaction.to_dict()
    assert data["sample_rate"] == 0.43
    assert elasticapm_client.events[constants.SPAN][0]["sample_rate"] == 0.43


def test_sample_rate_undefined_by_parent(elasticapm_client):
    trace_parent = TraceParent.from_string("00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03")
    elasticapm_client.begin_transaction("test", trace_parent=trace_parent)
    with elasticapm.capture_span("x"):
        pass
    transaction = elasticapm_client.end_transaction("test", "OK")
    data = transaction.to_dict()
    assert "sample_rate" not in data
    assert "sample_rate" not in elasticapm_client.events[constants.SPAN][0]


def test_trace_parent_not_set(elasticapm_client):
    elasticapm_client.begin_transaction("test", trace_parent=None)
    transaction = elasticapm_client.end_transaction("test", "OK")
    data = transaction.to_dict()
    assert data["trace_id"] is not None
    assert "parent_id" not in data


def test_ensure_parent_sets_new_id(elasticapm_client):
    transaction = elasticapm_client.begin_transaction("test", trace_parent=None)
    assert transaction.id == transaction.trace_parent.span_id
    span_id = transaction.ensure_parent_id()
    assert span_id == transaction.trace_parent.span_id


def test_ensure_parent_doesnt_change_existing_id(elasticapm_client):
    transaction = elasticapm_client.begin_transaction("test", trace_parent=None)
    assert transaction.id == transaction.trace_parent.span_id
    span_id = transaction.ensure_parent_id()
    span_id_2 = transaction.ensure_parent_id()
    assert span_id == span_id_2


def test_backdating_transaction(elasticapm_client):
    elasticapm_client.begin_transaction("test", start=time.time() - 1)
    elasticapm_client.end_transaction()
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    assert 900 < transaction["duration"] < 2000


def test_transaction_span_links(elasticapm_client):
    tp1 = TraceParent.from_string("00-aabbccddeeff00112233445566778899-0011223344556677-01")
    tp2 = TraceParent.from_string("00-00112233445566778899aabbccddeeff-aabbccddeeff0011-01")
    elasticapm_client.begin_transaction("a", links=[tp1, tp2])
    with elasticapm.capture_span("a", links=(tp1, tp2)):
        pass
    elasticapm_client.end_transaction("foo")
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    span = elasticapm_client.events[constants.SPAN][0]
    assert transaction["links"][0]["trace_id"] == "aabbccddeeff00112233445566778899"
    assert transaction["links"][0]["span_id"] == "0011223344556677"
    assert transaction["links"][1]["trace_id"] == "00112233445566778899aabbccddeeff"
    assert transaction["links"][1]["span_id"] == "aabbccddeeff0011"

    assert span["links"][0]["trace_id"] == "aabbccddeeff00112233445566778899"
    assert span["links"][0]["span_id"] == "0011223344556677"
    assert span["links"][1]["trace_id"] == "00112233445566778899aabbccddeeff"
    assert span["links"][1]["span_id"] == "aabbccddeeff0011"


def test_transaction_trace_continuation_continue(elasticapm_client):
    elasticapm_client.config.update("1", trace_continuation_strategy=constants.TRACE_CONTINUATION_STRATEGY.CONTINUE)
    tp = TraceParent.from_string("00-aabbccddeeff00112233445566778899-0011223344556677-01")
    elasticapm_client.begin_transaction("a", trace_parent=tp)
    elasticapm_client.end_transaction("foo")
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    assert transaction["trace_id"] == tp.trace_id
    assert "links" not in transaction


def test_transaction_trace_continuation_restart(elasticapm_client):
    elasticapm_client.config.update("1", trace_continuation_strategy=constants.TRACE_CONTINUATION_STRATEGY.RESTART)
    tp = TraceParent.from_string("00-aabbccddeeff00112233445566778899-0011223344556677-01")
    elasticapm_client.begin_transaction("a", trace_parent=tp)
    elasticapm_client.end_transaction("foo")
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    assert transaction["trace_id"] != tp.trace_id
    assert transaction["links"][0]["trace_id"] == tp.trace_id
    assert transaction["links"][0]["span_id"] == tp.span_id


def test_transaction_trace_continuation_restart_external(elasticapm_client):
    elasticapm_client.config.update(
        "1", trace_continuation_strategy=constants.TRACE_CONTINUATION_STRATEGY.RESTART_EXTERNAL
    )
    tp = TraceParent.from_string("00-aabbccddeeff00112233445566778899-0011223344556677-01")
    elasticapm_client.begin_transaction("a", trace_parent=tp)
    elasticapm_client.end_transaction("foo")
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    assert transaction["trace_id"] != tp.trace_id
    assert transaction["links"][0]["trace_id"] == tp.trace_id
    assert transaction["links"][0]["span_id"] == tp.span_id

    tp.add_tracestate("foo", "bar")
    elasticapm_client.begin_transaction("a", trace_parent=tp)
    elasticapm_client.end_transaction("foo")
    transaction = elasticapm_client.events[constants.TRANSACTION][1]
    assert transaction["trace_id"] == tp.trace_id
    assert "links" not in transaction



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import mock
import pytest


@mock.patch("tests.fixtures.DummyTransport.get_config")
def test_update_config(mock_get_config, elasticapm_client):
    assert elasticapm_client.config.transaction_sample_rate == 1.0
    assert elasticapm_client.config.config_version is None
    mock_get_config.return_value = 2, {"transaction_sample_rate": 0.1}, 30
    elasticapm_client.config.update_config()
    assert elasticapm_client.config.transaction_sample_rate == 0.1
    assert elasticapm_client.config.config_version == 2


@mock.patch("tests.fixtures.DummyTransport.get_config")
def test_environment_doesnt_override_central_config(mock_get_config, elasticapm_client):
    assert elasticapm_client.config.transaction_sample_rate == 1.0
    assert elasticapm_client.config.config_version is None
    mock_get_config.return_value = 2, {"transaction_sample_rate": 0.1}, 30
    with mock.patch.dict("os.environ", {"ELASTIC_APM_TRANSACTION_SAMPLE_RATE": "0.5"}):
        elasticapm_client.config.update_config()
    assert elasticapm_client.config.transaction_sample_rate == 0.1
    assert elasticapm_client.config.config_version == 2


@pytest.mark.parametrize("elasticapm_client", [{"transaction_sample_rate": 0.9}], indirect=True)
@mock.patch("tests.fixtures.DummyTransport.get_config")
def test_reset_to_original(mock_get_config, elasticapm_client):
    assert elasticapm_client.config.transaction_sample_rate == 0.9
    assert elasticapm_client.config.config_version is None
    assert not elasticapm_client.config.changed
    mock_get_config.return_value = 2, {"transaction_sample_rate": 0.1}, 30
    elasticapm_client.config.update_config()
    assert elasticapm_client.config.changed
    assert elasticapm_client.config.transaction_sample_rate == 0.1
    mock_get_config.return_value = 3, {}, 30
    elasticapm_client.config.update_config()
    assert not elasticapm_client.config.changed
    assert elasticapm_client.config.transaction_sample_rate == 0.9


@pytest.mark.parametrize("elasticapm_client", [{"transaction_sample_rate": 0.9}], indirect=True)
@mock.patch("tests.fixtures.DummyTransport.get_config")
def test_no_reset_if_version_matches(mock_get_config, elasticapm_client):
    assert elasticapm_client.config.transaction_sample_rate == 0.9
    assert elasticapm_client.config.config_version is None
    assert not elasticapm_client.config.changed
    mock_get_config.return_value = 2, {"transaction_sample_rate": 0.1}, 30
    elasticapm_client.config.update_config()
    assert elasticapm_client.config.changed
    assert elasticapm_client.config.transaction_sample_rate == 0.1
    mock_get_config.return_value = 2, {}, 30
    elasticapm_client.config.update_config()
    assert elasticapm_client.config.changed
    assert elasticapm_client.config.config_version == 2


@pytest.mark.parametrize("elasticapm_client", [{"central_config": False}], indirect=True)
def test_disable_central_config(elasticapm_client):
    assert elasticapm_client._config_updater is None


@mock.patch("tests.fixtures.DummyTransport.get_config")
def test_erroneous_config_is_ignored(mock_get_config, elasticapm_client):
    assert elasticapm_client.config.transaction_sample_rate == 1.0
    assert elasticapm_client.config.config_version is None
    mock_get_config.return_value = 2, {"transaction_sample_rate": "x"}, 30
    elasticapm_client.config.update_config()
    assert elasticapm_client.config.transaction_sample_rate == 1.0
    assert elasticapm_client.config.config_version == None
